.g-annotation-list-header
  margin-top 20px
  background-color #f0f0f0
  padding 5px 6px
  color #555
  font-size 16px
  font-weight bold

.g-annotation-list
  &.table
    width initial
    max-width initial
    min-width 100%

  table-layout fixed

  td
    white-space nowrap

  .g-annotation-toggle, .g-annotation-select
    width 30px

  .g-annotation-actions
    width 100px
    text-align right
